#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys, os, locale, re, urllib, socket
#import xml.etree.ElementTree as ET
from datetime import datetime
from sigil_bs4 import BeautifulSoup
from lxml import etree
from opf import _parse_xml_metadata, Metadata, XMLNS_DC, XMLNS_OPF, XMLNS_DCTERMS, XMLNS_XSI, XMLNS_CALIBRE
from xml.dom.minidom import parseString
from xml.dom import expatbuilder
import locale

"""
# this plugin requires PyQt5 which only comes with Sigil 0.9.8 and higher
try:
    from PyQt5 import uic, QtWidgets
    from PyQt5.QtCore import QTranslator, QAbstractTableModel, QPersistentModelIndex, QModelIndex, Qt, QDate
    from PyQt5.Qt import Qt, QMessageBox
    from PyQt5.QtGui import QStandardItemModel, QStandardItem, QKeyEvent, QIcon
    from PyQt5.QtWidgets import QApplication, QHeaderView, QDialog, QComboBox, QDateEdit, QStatusBar, QLabel, QLineEdit, QTextEdit
except Exception as e:
    print('Fatal error on line {}: '.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e, '\n\nThis plugin requires at least Sigil 0.9.8.\n\nClick OK to close.')
    sys.exit()
"""
from plugin_utils import Qt, QtCore, QtGui, QtWidgets
from plugin_utils import loadUi, PluginApplication, iswindows

# required for loading the .ui and .qm files from the plugin folder
plugin_dir = os.path.dirname(os.path.abspath(__file__))

class AddMetadata(QtWidgets.QDialog):

    def __init__(self, bk, parent = None):
        QtWidgets.QDialog.__init__(self, parent)
        self.bk = bk
        loadUi(os.path.join(plugin_dir, 'AddMetadata.ui'), self)

class MetaEditor(QtWidgets.QDialog):

    def __init__(self, bk):
        # used to detect updated data
        self.DirtyFlag = False
        
        #====================
        # Set up dialog box
        #====================
        plugin_dir = os.path.join(bk._w.plugin_dir, bk._w.plugin_name)
        #QtWidgets.QDialog.__init__(self, None)
        super(MetaEditor, self).__init__()
        loadUi(os.path.join(plugin_dir, 'MetaEditor.ui'), self)
        textEdit = QtWidgets.QTextEdit()
        self.resize(977, 627)
        
        #======================
        # get translated strings
        #======================
        self.bk = bk
        sigil_lang = bk.sigil_ui_lang[0:2]
        
        if sigil_lang not in ['en', 'fr', 'it', 'de', 'es']:
            sigil_lang = 'en'
        try:
            locale.setlocale(locale.LC_ALL, sigil_lang)
        except:
            #print('*** Locale not supported:',  sigil_lang, '***')
            pass
        
        from metadata import GetLanguageCodeMap, GetBasicMetadataNameDescription, GetBasicMetadataNameCode, GetMARCRelatorNameDescription, GetMARCRelatorNameCode, GetHeader, GetDataModifiedMessage

        # get language codes
        self.LanguageCodeMap = GetLanguageCodeMap(plugin_dir, sigil_lang)
        self.LanguageNameMap = {v: k for k, v in self.LanguageCodeMap.items()} 
        self.LanguageNameList = sorted(list(self.LanguageCodeMap.values()), key=locale.strxfrm)
        self.LanguageCodeList = list(self.LanguageCodeMap.keys())
 
        # get basic metadata strings
        self.BasicMetadataNameDescription = GetBasicMetadataNameDescription(plugin_dir, sigil_lang)
        self.BasicMetadataNameCode = GetBasicMetadataNameCode(plugin_dir, sigil_lang)
        self.BasicMetadataCodeName = {v: k for k, v in self.BasicMetadataNameCode.items()} 
        self.BasicMetadataNameList = sorted(list(self.BasicMetadataNameDescription.keys()), key=locale.strxfrm)
        self.BasicMetadataCodeList = sorted(list(self.BasicMetadataCodeName.keys()))
            
        # get MARC relators
        self.MARCRelatorNameDescription = GetMARCRelatorNameDescription(plugin_dir, sigil_lang)
        self.MARCRelatorNameCode = GetMARCRelatorNameCode(plugin_dir, sigil_lang) 
        self.MARCRelatorCodeName = {v: k for k, v in self.MARCRelatorNameCode.items()}
        self.MARCRelatorNameList = sorted(list(self.MARCRelatorNameDescription.keys()), key=locale.strxfrm)
        self.MARCRelatorCodeList = sorted(list(self.MARCRelatorCodeName.keys()))
        #if (len(self.MARCRelatorCodeList)) != 210:
            #print('Localizaton Error!')
            #return -1
        
        # get table view header
        self.header = GetHeader(plugin_dir, sigil_lang) 
        self.Creator = self.MARCRelatorCodeName['cre']
        self.Contributor = self.MARCRelatorCodeName['ctb']
        self.CreatorOrContributor = [self.Contributor, self.Creator ]
        
        # get metadata modified message
        self.DataModifiedMessage = GetDataModifiedMessage(plugin_dir, sigil_lang)
        
        #--------------------------------
        # check preferences
        #--------------------------------
        prefs = bk.getPrefs()
        self.custom_languages = None
        self.custom_roles = None
        if prefs != {}:
            #print('Preferences:', prefs, '\n')
            if 'languages' in prefs:
                self.custom_languages = []
                for lang in prefs['languages'].split(','):
                    lang = lang.strip()
                    if lang in self.LanguageCodeList:
                        self.custom_languages.append(self.LanguageCodeMap[lang])
                if self.custom_languages != []:
                    print('Custom languages:', self.custom_languages)
            else:
                print('No custom languages found.')
            if 'roles' in prefs:
                self.custom_roles = []
                for role in prefs['roles'].split(','):
                    role = role.strip()
                    if role in self.MARCRelatorCodeList:
                        self.custom_roles.append( self.MARCRelatorCodeName[role])
                if self.custom_roles != []:
                    print('Custom roles:', self.custom_roles, '\n')
            else:
                print('No custom roles found.')
        else:
            prefs['languages'] = ''
            prefs['roles'] = ''
            bk.savePrefs(prefs)
            print('No custom preferences found.')
        
        #==========================================
        # Get original metadata => ReadMetadataFromBook()
        #==========================================
        self.metadata_soup = None
        
        #-----------------------------------
        # parse metadata
        #-----------------------------------
        try:
            # undocumented feature:  https://stackoverflow.com/questions/24720668/python-read-data-from-xml-file
            dom = expatbuilder.parseString(bk.get_opf())
        except:
            print('Invalid metadata section!')
            return -1
            
        self.metadata = _parse_xml_metadata(dom)
        self.original_metadata = self.metadata.as_xml_element().toprettyxml(indent="  ")
        print('Original Metadata:\n\n', str(self.original_metadata))
        
        #--------------------------------------------------------
        # get unique identifier (mustn't be deleted)
        #--------------------------------------------------------
        self.unique_identifier_metadata = None
        # get the unique identifier 
        packagetag_soup = BeautifulSoup(bk.getpackagetag(), 'xml') 
        unique_identifier = packagetag_soup.package['unique-identifier']
        # search all ids
        if len(self.metadata.identifiers) >= 1:
            for i in range(0, len(self.metadata.identifiers)):
                contents, identifier, scheme = self.metadata.identifiers[i]
                if identifier == unique_identifier:
                    self.unique_identifier_metadata = (contents, identifier, scheme)
                    break 
        
        if self.unique_identifier_metadata == None:
            # try to get the uuid from the ncx file
            ncx_soup = BeautifulSoup(bk.readfile(bk.gettocid()), 'xml') 
            ncx_uuid_id = ncx_soup.find('meta', {'name' : 'dtb:uid' })
            if ncx_uuid_id is not None:
                contents = ncx_uuid_id['content']
                identifier = unique_identifier
                scheme = 'uuid'
                self.unique_identifier_metadata = (contents, identifier, scheme)
                self.metadata.add_identifier(contents, identifier, scheme)
                self.SetDirtyFlag()
                print('\n*** Broken unique identifier fixed: ' + contents, '***\n')
            else:
                print('Unique identifier not found.')
                return -1
        
        #=====================================
        # Populate GUI controls
        #=====================================
        
        #---------------------------------
        # LineEdits
        #---------------------------------
        
        # get TITLE (assume first title is the main title)
        self.original_title = ''
        if len(self.metadata.titles) >= 1:
            title, id, lang = self.metadata.titles[0]
            self.original_title = title
            self.original_title_lang = lang
            self.original_title_id = id
            # update TITLE LINE EDIT box
            self.leTitle.setText(title)
            #print('Original title:', title)
        else:
            print('Main title not found.')
        
        # get AUTHOR (assume first author is the main author)
        self.original_author = ''
        if len(self.metadata.creators) > 1:
            for i in range(0, len(self.metadata.creators)):
                self.original_author = self.metadata.creators[i]
                name, role, file_as, id, lang = self.metadata.creators[i]
                if role == 'aut':
                    # update AUTHOR LINE EDIT boxes
                    self.leAuthor.setText(name)
                    self.leAuthorFileAs.setText(file_as)
                    #print('Original author:', name)
                    #print('Original author sort as:', file_as)
                    if id != '':
                        #print('Original author id:', id)
                        pass
                    if lang != '':
                        #print('Original author lang:', lang)
                        pass
                    break
            # multiple dc:creator entries without 'aut' attribute; default to first dc:creator
            if self.leAuthor.text() == '':
                # assume first creator is the author
                self.original_author = self.metadata.creators[0]
                name, role, file_as, id, lang = self.metadata.creators[0]
                # update AUTHOR LINE EDIT boxes
                self.leAuthor.setText(name)
                self.leAuthorFileAs.setText(file_as)
                    
        elif len(self.metadata.creators) == 1:
            # assume first creator is the author
            self.original_author = self.metadata.creators[0]
            name, role, file_as, id, lang = self.metadata.creators[0]
            # update AUTHOR LINE EDIT boxes
            self.leAuthor.setText(name)
            self.leAuthorFileAs.setText(file_as)
        else:
            name = role = file_as = id = lang = None
            self.original_author = (name, role, file_as, id, lang)
            
        #print('Original author:', name)
        #print('Original author sort as:', file_as)
        if id != '':
            #print('Original author id:', id)
            pass
        if lang != '':
            #print('Original author lang:', lang)
            pass
            
        # get DESCRIPTION (assume first description is the main description)
        self.original_description = ''
        self.original_description_text = ''
        self.original_description_html = ''
        self.is_plain_description = False
        if len(self.metadata.description) >= 1:
            self.original_description = self.metadata.description[0]
            description, id, lang = self.metadata.description[0]
            self.original_description_text = description
            # test for html descriptions
            if '<' in description or '&' in description:
                self.is_plain_description = False
                self.teDescription.setHtml(description)
                self.original_description_html = self.teDescription.toHtml()
            else:
                self.is_plain_description = True
                self.teDescription.setPlainText(description)
            #print('Original description:', description)
        else:
            print('Main description not found.')        
            description = id = lang = None
            self.original_description = (description, id, lang)

        # connect changes to DirtyFlag
        self.leTitle.textChanged.connect(self.SetDirtyFlag)
        self.leAuthor.textChanged.connect(self.SetDirtyFlag)
        self.leAuthor.editingFinished.connect(self.LastFirst)
        self.leAuthorFileAs.textChanged.connect(self.SetDirtyFlag)
        self.teDescription.textChanged.connect(self.SetDirtyFlag)
        
        #-----------------------------------------------------------
        # Language QComboBox => SetLanguage()
        #-----------------------------------------------------------
        
        # get LANGUAGE
        self.main_language = sigil_lang
        if len(self.metadata.languages) >= 1:
            lang, id = self.metadata.languages[0]
            self.main_language_id = id
            # change En to en
            if len(lang) == 2 and lang.lower() != lang:
                lang = lang.lower()
                self.metadata.languages[0] = (lang, id)
                self.SetDirtyFlag()
                print('\n*** Invalid language code changed to:', lang, '***\n')
                
            # change 3 letter language codes with existing two letter codes to two letter language codes
            if len(lang) == 3 and lang in ['aar', 'abk', 'afr', 'aka', 'amh', 'ara', 'asm', 'ava', 'aym', 'aze', 'bak', 'bel', 'bis', 'bod', 'bre', 'cat', 'cha', 'cos', 'cre', 'cym', 'dan', 'deu', 'dzo', 'ell', 'eng', 'eus', 'fas', 'fin', 'fra', 'glg', 'guj', 'hau', 'heb', 'hin', 'hrv', 'hun', 'hye', 'iii', 'isl', 'ita', 'kat', 'kik', 'kor', 'kur', 'lat', 'lim', 'lub', 'mkd', 'msa', 'mya', 'nau', 'nde', 'nep', 'nld', 'nno', 'nor', 'nya', 'oci', 'oji', 'ori', 'oss', 'pan', 'que', 'ron', 'rus', 'san', 'sin', 'slv', 'smo', 'sna', 'som', 'sqi', 'srp', 'ssw', 'sun', 'swa', 'tam', 'tel', 'tgk', 'tha', 'tir', 'ton', 'tso', 'twi', 'ukr', 'urd', 'uzb', 'ven', 'vie', 'vol', 'wol', 'xho', 'yid', 'yor', 'zho', 'zul']:
                lang = lang[0:2]
                self.metadata.languages[0] = (lang, id)
                self.SetDirtyFlag()
                print('\n*** 3 letter language code changed to:', lang, '***\n')

            # change invalid language code to main UI language
            if lang in self.LanguageCodeMap:
                self.main_language = lang
            else:
                print('\n*** Invalid language code changed to:', sigil_lang, '***\n')
                self.metadata.languages[0] = (sigil_lang, id)
                self.SetDirtyFlag()
        else:
            id = None
            self.main_language_id = id
            self.metadata.languages.append((sigil_lang, id))
            self.SetDirtyFlag()
            print('\n*** Missing language code added: ', sigil_lang, '***\n')
            
        self.main_language_name = self.LanguageCodeMap[self.main_language]
        #print('Original Language Code:', self.main_language)
        #print('Original Language Name:', self.main_language_name)
        if self.main_language_id != '':
            #print('Original Language ID:', self.main_language_id)
            pass

        if self.custom_languages is not None and self.custom_languages != []:
            if self.main_language in self.LanguageCodeList and self.LanguageCodeMap[self.main_language] not in self.custom_languages:
                self.custom_languages.append(self.LanguageCodeMap[self.main_language])
            self.cbLanguages.addItems(sorted(self.custom_languages, key=locale.strxfrm))
        else:
            self.cbLanguages.addItems(self.LanguageNameList)
        index = self.cbLanguages.findText(self.main_language_name, Qt.MatchFixedString)
        if index >= 0:
            self.cbLanguages.setCurrentIndex(index)
        self.cbLanguages.currentIndexChanged.connect(self.SetDirtyFlag)
            
        #==========================
        # Set up original QTableView data
        #==========================
        
        # define model
        self.MetaModel = QtGui.QStandardItemModel()
        self.MetaModel.setColumnCount(4)
        self.MetaModel.setHorizontalHeaderLabels(self.header)
        
        # initialize table view
        self.tvMetaTable.setModel(self.MetaModel)
        self.tvMetaTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.tvMetaTable.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.tvMetaTable.resizeColumnsToContents()
        
        #===========================
        # **** Add data to table view *****
        #===========================
        
        # DC:TITLE (skip first title = main title)
        for title in self.metadata.titles[1:]:
            title, id, lang = title
            name = self.BasicMetadataCodeName['title']
            value = title
            fileas = id
            role = lang
            row = self.MetaModel.rowCount()
            self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:CREATOR
        for creator in self.metadata.creators:
            if creator !=  self.original_author:
                name, role_code, fileas, id, lang = creator
                value = name
                if id != '' or lang != '':
                    print('\n*** Warning: creator/contributor id or language attribute ignored: ', creator, '***\n')
                if role_code != '' and role_code in self.MARCRelatorCodeList:
                    name = self.MARCRelatorCodeName[role_code]
                else:
                    name = self.MARCRelatorCodeName['cre']
                fileas = fileas
                role = self.MARCRelatorCodeName['cre']
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)
                
        # DC:CONTRIBUTOR
        for contributor in self.metadata.contributors:
            name, role_code, fileas, id, lang = contributor
            value = name
            if role_code != '' and role_code in self.MARCRelatorCodeList:
                name = self.MARCRelatorCodeName[role_code]
            else:
                name = self.MARCRelatorCodeName['ctb']
            fileas = fileas
            role = self.MARCRelatorCodeName['ctb']
            row = self.MetaModel.rowCount()
            self.AddMetadataToTable(row, name, value, fileas, role)
        
        # DC:SUBJECT
        for subject in self.metadata.subjects:
            sub, id, lang = subject
            name = self.BasicMetadataCodeName['subject']
            value = sub
            fileas = id
            role = lang
            if value != '':
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:LANGUAGE
        for language in self.metadata.languages:
            language_code, id = language
            if language_code != '': 
                if  language_code != self.main_language:
                    name = self.BasicMetadataCodeName['language']
                    if language_code in self.LanguageCodeList:
                        value = self.LanguageCodeMap[language_code]
                    else:
                        value = self.LanguageCodeMap[sigil_lang]
                    fileas = id
                    role = ''
                    row = self.MetaModel.rowCount()
                    self.AddMetadataToTable(row, name, value, fileas, role)
                    #if language_code in self.LanguageCodeList:
                        #print('Secondary language found: ', language_code)
                        #print('Secondary language name: ', self.LanguageCodeMap[language_code])
                    #else:
                        #print('Invalid language code found: ', language_code)
                
        # DC:IDENTIFIER:
        for identifier in self.metadata.identifiers:
            # skip the unique identifier that can't be edited
            if identifier !=  self.unique_identifier_metadata:
                contents, identifier, scheme = identifier
                if scheme in ['DOI', 'ISBN', 'ISSN', 'MOBI-ASIN']:
                    name = self.BasicMetadataCodeName[scheme]
                else:
                    name = self.BasicMetadataCodeName['customidentifier']
                value = contents
                fileas = scheme
                role = ''
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)
        
        # DC:DATE
        for date in self.metadata.dates:
            date, event, id = date
            if event == 'publication':
                name = self.BasicMetadataCodeName['publication']
            elif event == 'creation':
                name = self.BasicMetadataCodeName['creation']
            elif event == 'modification':
                name = self.BasicMetadataCodeName['modification']
            elif event == 'conversion':
                name = self.BasicMetadataCodeName['conversion']
            else:
                name = self.BasicMetadataCodeName['date']
            value = date
            fileas = id
            role = ''
            row = self.MetaModel.rowCount()
            self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:DESCRIPTION (skip first entry = main description)
        for description in self.metadata.description[1:]:
            desc, id, lang = description
            name = self.BasicMetadataCodeName['description']
            value = desc
            if value != '':
                fileas = id
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)
  
        # DC:PUBLISHER
        for publisher in self.metadata.publisher:
            pub, id, lang = publisher
            name = self.BasicMetadataCodeName['publisher']
            value = pub
            if value != '':
                fileas = id
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:TYPE
        for dc_type in self.metadata.dc_type:
            type, id = dc_type
            name = self.BasicMetadataCodeName['type']
            value = type
            if value != '':
                fileas = id 
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:RELATION
        for relation in self.metadata.relation:
            rel, id, lang = relation
            name = self.BasicMetadataCodeName['relation']
            value = rel
            if value != '':
                fileas = id 
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:FORMAT
        for format in self.metadata.dc_format:
            fmt, id = format
            name = self.BasicMetadataCodeName['format']
            value = fmt
            if value != '':
                fileas = id
                role = ''
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:SOURCE
        for source in self.metadata.source:
            src, id, lang = source
            name = self.BasicMetadataCodeName['source']
            value = src
            if value != '':
                fileas = id
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)
            
        # DC:COVERAGE
        for coverage in self.metadata.coverage:
            cvg, id, lang = coverage
            name = self.BasicMetadataCodeName['coverage']
            value = cvg
            if value != '':            
                fileas = id
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)

        # DC:RIGHTS
        for right in self.metadata.right:
            rt, id, lang = right
            name = self.BasicMetadataCodeName['rights']
            value = rt
            if value != '':
                fileas = id
                role = lang
                row = self.MetaModel.rowCount()
                self.AddMetadataToTable(row, name, value, fileas, role)
                
        # META
        for meta in self.metadata.metas:
            name, content, scheme, id, lang = meta
            value = content
            fileas = scheme
            if id != '' or lang != '':
                print('\n*** Warning: meta id or language attribute ignored: ', meta, '***\n')
            role = 'Meta'
            row = self.MetaModel.rowCount()
            self.AddMetadataToTable(row, name, value, fileas, role)
        
        # connect changes to dirty flag
        self.MetaModel.dataChanged.connect(self.SetDirtyFlag)
        self.tvMetaTable.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.tvMetaTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        #self.tvMetaTable.resizeColumnsToContents()
       
        #=============
        # Map buttons
        #=============
        
        # OK & Cancel buttons
        self.accepted.connect(lambda: self.MainOKButton(self.bk))
        self.rejected.connect(self.MainCancelButton)

        # Add Basic & Add Role buttons
        self.btAddBasic.clicked.connect(self.AddBasic)
        self.btAddRole.clicked.connect(self.AddRole)
        
        # Copy, Remove, Move Up and Move Down table buttons
        self.btCopy.clicked.connect(self.Copy)
        self.btRemove.clicked.connect(self.Remove)
        
        # Move Up and Move Down are disabled
        self.tbMoveUp.hide()
        self.tbMoveDown.hide()
        #self.tbMoveUp.clicked.connect(self.MoveUp)
        #self.tbMoveDown.clicked.connect(self.MoveDown)

        
    #=========================
    # ****define button actions ****
    #=========================
    def SetDirtyFlag(self):
        self.DirtyFlag = True
    
    def MainOKButton(self, bk):
        if self.DirtyFlag == False:
            print('\nNothing changed. Metadata NOT updated.')
            return
        
        self.new_author = self.leAuthor.text()
        self.new_author_fileas = self.leAuthorFileAs.text()
        self.new_title = self.leTitle.text()

        #===========================
        # DC: DESCRIPTION handler
        #===========================
        
        # check for plain text/html
        if self.is_plain_description:
            self.new_description = self.teDescription.toPlainText()
        else:
            if self.teDescription.toPlainText() != '':
                self.new_description_html = self.teDescription.toHtml()
                if self.new_description_html == self.original_description_html:
                    # use the original html code, if the description hasn't changed
                    self.new_description = self.original_description_text
                else:
                    # get rid of default Qt inline styles
                    description_soup = BeautifulSoup(self.new_description_html, 'html.parser')
                    for span in description_soup.find_all('span'):
                        if 'style' in span.attrs:
                            if 'font-size:xx-large; font-weight:600;' in span['style']:
                                span.attrs = {}
                                span.name = 'h1'
                            elif 'font-size:x-large; font-weight:600;' in span['style']:
                                span.attrs = {}
                                span.name = 'h2'
                            elif 'font-size:large; font-weight:600;' in span['style']:
                                span.attrs = {}
                                span.name = 'h3'
                            elif 'font-weight:600; font-style:italic;' in span['style']:
                                span.attrs = {}
                                span.name = 'em'
                                span.wrap(description_soup.new_tag('strong'))
                            elif 'font-weight:600;' in span['style']:
                                span.attrs = {}
                                span.name = 'strong'
                            elif 'font-style:italic;' in span['style']:
                                span.attrs = {}
                                span.name = 'em'
                            elif 'text-decoration: underline;' in span['style']:
                                span.attrs = {}
                                span.name = 'u'
                            elif 'text-decoration: line-through;' in span['style']:
                                span.attrs = {}
                                span.name = 's'
                            elif 'vertical-align:super;' in span['style']:
                                span.attrs = {}
                                span.name = 'sup'
                            elif 'vertical-align:sub;' in span['style']:
                                span.attrs = {}
                                span.name = 'sub'
                            else:
                                span.unwrap()
                                
                    self.new_description = ''
                    for content in description_soup.body.contents:
                        content.attrs = {} 
                        self.new_description += str(content)
                    self.new_description = self.new_description.strip()
            else:
                self.new_description = ''
        
        #---------------------------------
        # get new LANGUAGE
        #------------------------------------
        self.new_language_name = self.cbLanguages.currentText()
        self.new_language_code = self.LanguageNameMap[self.new_language_name]

        #print('\nNew title:', self.new_title)
        #print('New author:', self.new_author)
        #print('New author sort as:', self.new_author_fileas)
        #print('New description:', self.new_description)
        #print('New language code:', self.new_language_code)
        #print('New Langauge name:', self.new_language_name)

        #***********************************************
        # ****** rebuild metadata section ******
        #***********************************************
        
        # create empty metadata section
        new_metadata = Metadata()
        
        #----------------------------------
        # 1. Non-tableview items
        #----------------------------------
        
        # add main LANGUAGE
        new_metadata.add_language(self.new_language_code, self.main_language_id)
        
        # add main TITLE
        new_metadata.add_title(self.new_title, self.original_title_id, self.original_title_lang)

        # add main AUTHOR
        if self.new_author != '' and self.new_author_fileas != '':
            # get original author id and lang values
            dummy, dummy, dummy, id, lang = self.original_author
            # get new name and fileas
            name = self.new_author
            file_as = self.new_author_fileas
            new_metadata.add_creator(name, 'aut', file_as, id, lang)
        else:
            if self.new_author != '':
                # get original author id and lang values
                dummy, dummy, dummy, id, lang = self.original_author
                # get new name 
                name = self.new_author
                file_as = None
                new_metadata.add_creator(name, 'aut', file_as, id, lang)
                
        # add DESCRIPTION
        if self.new_description != '':
             # get original description id and lang values
            dummy, id, lang = self.original_description
            description = self.new_description
            new_metadata.add_description(description, id, lang)
        
        # UNIQUE IDENTIFIER
        contents, identifier, scheme = self.unique_identifier_metadata
        new_metadata.add_identifier(contents, identifier, scheme)

        #============================
        # process QTableView entries
        #============================
        
        # look for DC:CREATOR entries
        row = self.MetaModel.rowCount()
        for current_row in range(0, row):
            Name = self.MetaModel.data(self.MetaModel.index(current_row, 0))
            Value = self.MetaModel.data(self.MetaModel.index(current_row, 1))
            FileAs = self.MetaModel.data(self.MetaModel.index(current_row, 2))
            Role = self.MetaModel.data(self.MetaModel.index(current_row, 3))
        
            # DC entries
            if Name in self.BasicMetadataNameList:
                code = self.BasicMetadataNameCode[Name]
                # TITLE
                if code == 'title':
                    #Name, Value, FileAs, Role
                    title = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_title(title, id, lang)
                # DATE
                elif code == 'creation':
                    date = Value
                    id = FileAs
                    event = 'creation'
                    new_metadata.add_date(date, event, id)
                elif code == 'publication':
                    date = Value
                    id = FileAs                    
                    event = 'publication'
                    new_metadata.add_date(date, event, id)
                elif code == 'modification':
                    date = Value
                    id = FileAs
                    event = 'modification'
                    new_metadata.add_date(date, event, id)
                elif code == 'conversion':
                    date = Value
                    id = FileAs
                    event = 'conversion'
                    new_metadata.add_date(date, event, id)
                elif code == 'date':
                    date = Value
                    id = FileAs
                    event = None
                    new_metadata.add_date(date, event, id)
                # LANGUAGE
                elif code == 'language':
                    lang = self.LanguageNameMap[Value]
                    id = FileAs
                    new_metadata.add_language(lang, id)
                # DESCRIPTION
                elif code == 'description':
                    description = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_description(description, id, lang)
                # PUBLISHER
                elif code == 'publisher':
                    publisher = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_publisher(publisher, id, lang)
                # TYPE
                elif code == 'type':
                    type = Value
                    id = FileAs
                    new_metadata.add_dc_type(type, id)
                # FORMAT
                elif code == 'format':
                    format = Value
                    id = FileAs
                    new_metadata.add_format(format, id)
                # SOURCE
                elif code == 'source':
                    source = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_source(source, id, lang)
                # RELATION
                elif code == 'relation':
                    relation = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_relation(relation, id, lang)
                # COVERAGE
                elif code == 'coverage':
                    coverage = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_coverage(coverage, id, lang)
                # RIGHT
                elif code == 'rights':
                    right = Value
                    id = FileAs
                    lang = Role
                    new_metadata.add_right(right, id, lang)
                # SUBJECT
                elif code == 'subject':
                    subjects = Value.split(';')
                    id = FileAs
                    lang = Role
                    for subject in subjects:
                        new_metadata.add_subject(subject.strip(), id, lang)
                       
                # predefined id values
                elif code in ['DOI', 'ISBN', 'ISSN', 'MOBI-ASIN', 'customidentifier']:
                    content = Value
                    id = None
                    scheme = FileAs
                    new_metadata.add_identifier(content, id, scheme)
                else:
                    print('*********** Unknown DC: entry found !!! ***************', Name, code)

            #--------------------------
            # MARC Relators
            #--------------------------
            elif Role in self.MARCRelatorNameList:
                code = self.MARCRelatorNameCode[Role]
                if code == 'cre' or 'ctb':
                    name = Value
                    relators = name.split(';')
                    for relator in relators:
                        name = relator.strip()
                        role = self.MARCRelatorNameCode[Name]
                        fileas = FileAs
                        id = None
                        lang = None
                        if FileAs != '':
                            if code == 'cre':
                                new_metadata.add_creator(name, role, fileas, id, lang)
                            else:
                                new_metadata.add_contributor(name, role, fileas, id, lang)
                        else:
                            if code == 'cre':
                                new_metadata.add_creator(name, role, fileas, id, lang)
                            else:
                                new_metadata.add_contributor(name, role, fileas, id, lang)
                else:
                    print('*************** Unknown MARC relator found!!! **********************', Role)

            #----------------------
            # META tags
            #----------------------
            elif Role == 'Meta':
                name = Name
                content = Value
                scheme = FileAs
                id = None
                lang = None
                new_metadata.add_meta(name, content, scheme, id, lang)

        #=======================
        # write updated metadata
        #=======================

        print('\nUpdated metadata:\n\n', new_metadata.as_xml_element().toprettyxml(indent="  "))        
        updated_metadata = new_metadata.as_xml_element().toprettyxml(indent="  ")
        bk.setmetadataxml(updated_metadata)
        print('\nMetadata updated.')
        
    def MainCancelButton(self):
        # check for metadata changes and warn the user
        if self.DirtyFlag == True:
                result =  QtWidgets.QMessageBox.warning(self, "MetadataEditorPlugin", self.DataModifiedMessage, QtWidgets.QMessageBox.Discard | QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.Cancel, QtWidgets.QMessageBox.Discard)
                if result == QtWidgets.QMessageBox.Save:
                    self.MainOKButton(self.bk)
                    print('\nMetadata updated.')
                else:
                    print('\nMetadata NOT updated.')
        else:
            print('\nMetadata NOT updated.')
    
    # copy selected row        
    def Copy(self):
        
        if self.tvMetaTable.selectionModel().hasSelection():
            indexes = self.tvMetaTable.selectionModel().selectedRows() 
            #select first row only
            index = sorted(indexes)[0]
            current_row = index.row()
            Name = self.MetaModel.data(self.MetaModel.index(current_row, 0))
            Value = self.MetaModel.data(self.MetaModel.index(current_row, 1))
            FileAs = self.MetaModel.data(self.MetaModel.index(current_row, 2))
            Role = self.MetaModel.data(self.MetaModel.index(current_row, 3))
            self.AddMetadataToTable(current_row, Name, Value, FileAs, Role)
            self.tvMetaTable.resizeColumnsToContents()
            self.tvMetaTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
            self.tvMetaTable.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
            self.SetDirtyFlag()
        else:
            print('No row selected!')

    # delete selected row(s)
    def Remove(self):
        try:
            if self.tvMetaTable.selectionModel().hasSelection():
                indexes =[QtCore.QPersistentModelIndex(index) for index in self.tvMetaTable.selectionModel().selectedRows()]
                maxrow = max(indexes, key=lambda x: x.row()).row()
                next_ix = QtCore.QPersistentModelIndex(self.MetaModel.index(maxrow+1, 0))
                for index in indexes:
                    self.MetaModel.removeRow(index.row())
                self.tvMetaTable.setCurrentIndex(QtCore.QModelIndex(next_ix))
                self.SetDirtyFlag()
            else:
                print('No row selected!')
        except Exception as e:
            print('Fatal error on line {}: '.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
                
    # allows you to delete rows 
    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Backspace, Qt.Key_Delete):
            self.Remove()
        QtWidgets.QDialog.keyPressEvent(self, event)
            
    # move row up
    #def MoveUp(self):
        #print('MoveUp: not implemented.')

    # move row down
    #def MoveDown(self):
        #print('MoveDown: not implemented.')

    #========================================
    # ******* add "basic metadata" i.e. dc: entries ******
    #========================================
    def AddBasic(self):
        self.AddBasicDialog = AddMetadata(self)
        self.AddBasicDialog.accepted.connect(self.AddBasicOKButton)
        self.AddBasicDialog.rejected.connect(self.AddBasicCancelButton)
        self.AddBasicDialog.setWindowModality(Qt.ApplicationModal)
        self.AddBasicDialog.lwProperties.clear()
        self.AddBasicDialog.lwProperties.addItems(self.BasicMetadataNameList)
        self.AddBasicDialog.lbDescription.setText(self.BasicMetadataNameDescription[self.BasicMetadataNameList[0]])
        self.AddBasicDialog.lwProperties.itemSelectionChanged.connect(self.UpdateBasicDescription)
        self.AddBasicDialog.show()

    def UpdateBasicDescription(self):
        currentItem = self.AddBasicDialog.lwProperties.currentItem().text()
        if currentItem in self.BasicMetadataNameList:
            self.AddBasicDialog.lbDescription.clear()
            self.AddBasicDialog.lbDescription.setText(self.BasicMetadataNameDescription[currentItem])

    def AddBasicOKButton(self):
        fullname = self.AddBasicDialog.lwProperties.currentItem().text()
        fullname_code = self.BasicMetadataNameCode[fullname]
        row = self.MetaModel.rowCount()
        # add date stamp to all dates
        if fullname_code in ['publication', 'creation', 'modification', 'conversion']:
            date = datetime.today().strftime('%Y-%m-%d')
            self.AddMetadataToTable(row, fullname, date, '', '')
        elif fullname_code in ['DOI', 'ISBN', 'ISSN', 'MOBI-ASIN', 'customidentifier']:
            self.AddMetadataToTable(row, fullname, '', fullname_code, '')
        elif fullname_code in ['calibre:series', 'calibre:series_index', 'calibre:title_sort']:
            self.AddMetadataToTable(row, fullname_code, '', '', 'Meta')
        else:
            self.AddMetadataToTable(row, fullname, '', '', '')

    #===========================
    # ********** populate table ***********
    #===========================
    def AddMetadataToTable(self, row, name, value, fileas, role):
        self.MetaModel.insertRow(row)
        self.MetaModel.setData(self.MetaModel.index(row, 0), name)
        self.MetaModel.item(row, 0).setEditable(False)
        
        #-------------------------------------------------
        # add date widget to <dc:date> entry
        #-------------------------------------------------
        if name in self.BasicMetadataNameList and self.BasicMetadataNameCode[name] in ["publication", "creation", "modification", "conversion", "date"]:
            if value is not None and value != '':
                displayFormat = 'yyyy-MM-dd'
                if len(value) == 4:
                    displayFormat = 'yyyy'
                    #value += '-01-01'
                elif len(value) > 10:
                    value = value[0:10]

                self.MetaModel.setData(self.MetaModel.index(row, 1), value)
            else:
                today = QtCore.QDate.currentDate().toString(Qt.ISODate)
                self.MetaModel.setData(self.MetaModel.index(row, 1), today)
                displayFormat = 'yyyy-MM-dd'
            item = self.tvMetaTable.model().item(row, 1)
            date = self.MetaModel.data(self.MetaModel.index(row, 1))
            deDelegate = QtWidgets.QDateEdit()
            #deDelegate.setCalendarPopup(True)
            deDelegate.setDisplayFormat(displayFormat)
            deDelegate.dateChanged.connect(lambda date, it=item: it.setText(date.toString(displayFormat)))
            myQDate = QtCore.QDate.fromString(date, displayFormat)
            deDelegate.setDate(myQDate)
            self.tvMetaTable.setIndexWidget(item.index(), deDelegate)
            
        self.MetaModel.setData(self.MetaModel.index(row, 1), value)
        
        #-------------------------------------------------------------------
        # add language combox for <dc:language> entries
        #-------------------------------------------------------------------
        if value in self.LanguageNameList or name == self.BasicMetadataCodeName['language']:
            item = self.tvMetaTable.model().item(row, 1)
            cbDelegate = QtWidgets.QComboBox()
            cbDelegate.currentIndexChanged[str].connect(item.setText)
            if self.custom_languages is not None and self.custom_languages != []:
                cbDelegate.addItems(sorted(self.custom_languages, key=locale.strxfrm))
            else:
                cbDelegate.addItems(self.LanguageNameList)
            self.tvMetaTable.setIndexWidget(item.index(), cbDelegate)
            index = cbDelegate.findText(value, Qt.MatchFixedString)
            if index >= 0:
                cbDelegate.setCurrentIndex(index)
                
        self.MetaModel.setData(self.MetaModel.index(row, 2), fileas)
        if role in self.CreatorOrContributor or fileas == 'customidentifier':
            self.MetaModel.item(row, 2).setEditable(True)
        else:
            self.MetaModel.item(row, 2).setEditable(False)
        self.MetaModel.setData(self.MetaModel.index(row, 3), role)
        self.MetaModel.item(row, 3).setEditable(True)
        
        #-------------------------------------------------------------
        # add <dc:creator>/<dc:contributor> combobox
        #--------------------------------------------------------------
        if role in self.CreatorOrContributor:
            item = self.tvMetaTable.model().item(row, 3)
            cbDelegate = QtWidgets.QComboBox()
            # the overloaded currentIndexChanged slot with a QString
            # parameter went away in Qt6. Use currentTextChanged instead.
            # cbDelegate.currentIndexChanged[str].connect(item.setText)
            cbDelegate.currentTextChanged.connect(item.setText)
            if role == self.Creator:
                cbDelegate.addItems([self.Creator, self.Contributor])
            else:
                cbDelegate.addItems([self.Contributor, self.Creator])
            self.tvMetaTable.setIndexWidget(item.index(), cbDelegate)
        else:
            self.MetaModel.item(row, 3).setEditable(False)
        self.tvMetaTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.tvMetaTable.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def LastFirst(self):
        author_parts = self.leAuthor.text().split(' ')
        if len(author_parts) >= 2 and ',' not in self.leAuthor.text():
            LastFirst = author_parts[len(author_parts) - 1] + ', ' + " ".join(author_parts[0:len(author_parts) - 1])
            if self.leAuthorFileAs.text() == '':
                self.leAuthorFileAs.setText(LastFirst)
        else:
            if self.leAuthorFileAs.text() == '':
                self.leAuthorFileAs.setText(self.leAuthor.text())
            
    def AddBasicCancelButton(self):
        pass

    #============================
    # add MARC roles
    #============================
    def AddRole(self):
        self.AddRoleDialog = AddMetadata(self)
        self.AddRoleDialog.accepted.connect(self.AddRoleOKButton)
        self.AddRoleDialog.rejected.connect(self.AddRoleCancelButton)
        self.AddRoleDialog.setWindowModality(Qt.ApplicationModal)
        self.AddRoleDialog.lwProperties.clear()
        if self.custom_roles is not None and self.custom_roles != []:
            self.AddRoleDialog.lwProperties.addItems(self.custom_roles)
        else:
            self.AddRoleDialog.lwProperties.addItems(self.MARCRelatorNameList)
        self.AddRoleDialog.lbDescription.setText(self.MARCRelatorNameDescription[self.MARCRelatorNameList[0]])
        self.AddRoleDialog.lwProperties.itemSelectionChanged.connect(self.UpdateRoleDescription)
        self.AddRoleDialog.show()
        
    def UpdateRoleDescription(self):
        currentItem = self.AddRoleDialog.lwProperties.currentItem().text()
        if currentItem in self.MARCRelatorNameList:
            self.AddRoleDialog.lbDescription.clear()
            self.AddRoleDialog.lbDescription.setText(self.MARCRelatorNameDescription[currentItem])

    def AddRoleOKButton(self):
        fullname = self.AddRoleDialog.lwProperties.currentItem().text()
        fullname_code = self.MARCRelatorNameCode[fullname]
        role = self.MARCRelatorCodeName['ctb']
        row = self.MetaModel.rowCount()
        self.AddMetadataToTable(row, fullname, '', '', role)
        
    def AddRoleCancelButton(self):
        pass


"""
def dark_palette(bk, app):
    supports_theming = (bk.launcher_version() >= 20200117)
    if not supports_theming:
        return
    if bk.colorMode() != "dark":
        return
    try:
        from PyQt5.QtGui import QColor, QPalette
        from PyQt5.QtWidgets import QStyleFactory
    except ImportError:
        return

    p = QPalette()
    sigil_colors = bk.color
    dark_color = QColor(sigil_colors("Window"))
    disabled_color = QColor(127,127,127)
    dark_link_color = QColor(108, 180, 238)
    text_color = QColor(sigil_colors("Text"))
    p.setColor(p.Window, dark_color)
    p.setColor(p.WindowText, text_color)
    p.setColor(p.Base, QColor(sigil_colors("Base")))
    p.setColor(p.AlternateBase, dark_color)
    p.setColor(p.ToolTipBase, dark_color)
    p.setColor(p.ToolTipText, text_color)
    p.setColor(p.Text, text_color)
    p.setColor(p.Disabled, p.Text, disabled_color)
    p.setColor(p.Button, dark_color)
    p.setColor(p.ButtonText, text_color)
    p.setColor(p.Disabled, p.ButtonText, disabled_color)
    p.setColor(p.BrightText, Qt.red)
    p.setColor(p.Link, dark_link_color)
    p.setColor(p.Highlight, QColor(sigil_colors("Highlight")))
    p.setColor(p.HighlightedText, QColor(sigil_colors("HighlightedText")))
    p.setColor(p.Disabled, p.HighlightedText, disabled_color)

    app.setStyle(QStyleFactory.create("Fusion"))
    app.setPalette(p)
"""

        
def run(bk):
    # define dialog box
    sigil_lang = bk.sigil_ui_lang[0:2]
    mdp = True if iswindows else False
    app_icon = os.path.join(plugin_dir, 'app.ico')
    app = PluginApplication(sys.argv, bk, app_icon=app_icon, match_dark_palette=mdp,
                            load_qtbase_translations=False, load_qtplugin_translations=False)
    # dark_palette(bk, app)

    if bk.epub_version() == "2.0":
        translator = QtCore.QTranslator()
        translator.load('dialogs_' + sigil_lang + '.qm', plugin_dir)
        app.installTranslator(translator)
        window = MetaEditor(bk)
        window.show()
        app.exec()
    else:
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setWindowIcon(QtGui.QIcon(os.path.join(plugin_dir, 'app.ico')))
        msg.setWindowTitle("Wrong epub version")
        msg.setText("This plugin only works with epub2 books.")
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        buttonClicked = msg.exec()
        print('Click OK to close.')
    
    return 0
    
def main():
    print('I reached main when I should not have\n')
    return -1

if __name__ == "__main__":
    sys.exit(main())
